/*
	NoQuaFra.cpp
	NoQuaFra - Non Quaternion 3d Fractal Generator
	Copyright (C)Billy Stiltner 2002
// gcc -o mandelbox mandelbox.cc -lX11 -lGL -lGLU
*/

//////////////////////////////////////////////////////////////////////
//#include "windows.h"




#include <stdio.h>
#include <stdlib.h>

#include<string.h>
#include<unistd.h>
#include<math.h>
#include<time.h>
#include<sys/time.h>
#include<X11/Xlib.h>
#include<X11/XKBlib.h>
#include<GL/glx.h>
#include<GL/glext.h>
#include<GL/glu.h>
#include <X11/keysymdef.h>

#include "noquafraaux.h"

#include "defaultmap.h"
#include "mandelbox.h"

//////////////////////////////////////////////////////////////////////

#define FRC_POINTS  0
#define FRC_CUBES   1
#define FRC_SPHERES 2
//////////////////////////////////////////////////////////////////////

GLboolean should_rotate;
float angle ;

float anglex ;
float angley ;
float anglez ;
mandelbrot mdel;

int colorcycle;
GLboolean should_colorcycle;
GLboolean rotatex;
GLboolean rotatey;
GLboolean rotatez;
float move_x,xpos;
float move_y,ypos;
float move_z,zpos;
int mouse_left,mouse_right;

int nsize;
int pointstyle;
int pointsize;
GLboolean uselighting;
GLboolean zooming;


//////////////////////////////////////////////////////////////////////////////////
Display                 *dpy;
Window                  root, win;
GLint                   att[]   = { GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None };
XVisualInfo             *vi;
GLXContext              glc;
Colormap                cmap;
XSetWindowAttributes    swa;
XWindowAttributes       wa;
XEvent                  xev;

float                   TimeCounter, LastFrameTimeCounter, DT, prevTime = 0.0, FPS;
struct timeval          tv, tv0;
int                     Frame = 1, FramesPerFPS;

GLfloat                 rotation_matrix[16];
float                   rot_z_vel = 50.0, rot_y_vel = 30.0;


//////////////////////////////////////////////////////////////////////

void quit_program( int code );
void handle_key_down( int xk );

void process_events( void );
void CreateWindow(int width, int height);

void draw_screen( void );
void setup_opengl( int width, int height );

void InitTimeCounter();

void UpdateTimeCounter();

void CalculateFPS();

/*
void MouseMotionHandler(Uint8 type, SDL_MouseMotionEvent event);
void MouseButtonHandler(Uint8 type, SDL_MouseButtonEvent event);
*/
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
int main( int argc, char* argv[] )
{

	//const SDL_VideoInfo* info = NULL;
	int width = 0;
	int height = 0;
	int bpp = 0;
	int flags = 0;

	pointstyle=FRC_POINTS;
	pointsize=10;
	nsize=96;

	should_rotate = GL_FALSE;

	uselighting=GL_FALSE;

	zooming = GL_FALSE;

	angle = 0.0f;
	anglex=0.0f;
	angley=0.0f;
	anglez=0.0f;

	colorcycle=0;
	move_x=0;
	move_y=0;
	move_z=0;

	//this moves the fractal so its center is 0,0,-50
	xpos=-(nsize/2);			 //set this to  - fractal cube width / 2   or move the fractal left half its width
	ypos=-(nsize/2);			 //set this to  - fractal cube height / 2  or move the fractal down half its height
	zpos=-(nsize*2);			 ////set this to  - fractal cube depth * 2 or move the fractal forward twice  its length

	

	width = 640;
	height = 480;

CreateWindow(width, height);
	

	setup_opengl( width, height );
 InitTimeCounter();

	setallcolors();
	//mdel.init(-2.5, 1.5, 2.5,-1.5, 0.0, 0.0, 4, 10000, 0, 0, 100,75);
	//mdel.init(-2.5, 1.5, 2.5,-1.5, 0.0, 0.0, 4, 1000, 0, 0, 100,75);

	//mdel.init(-1.1875, -1.1875, 1.25,1.25, 0.0, 0.0, 4, 1000, 0, 0, 100,100);

	// mdel.init(-1.75, -1.75, 0.0,0.0, 0.0, 0.0, 4, 1000, 0, 0, nsize,nsize);
/*	mdel.init(-1.75, -1.75, -1.75,
		1.75,  1.75,  1.75,
		-0.488193901,0.572800587,   0.0,
		0,     0,     0,
		nsize, nsize, nsize,
		1.75,
		1024,
		 64);*/

mdel.init(-3, -3, -3,
		3,  3,  3,
		0.0,0.0,   0.0,
		0,     0,     0,
		nsize, nsize, nsize,
		-2.618,
		1024,
		 64);

	mdel.iterate();
	mdel.setcube();
	mdel.setzoomcube();

	while( 1 )
	{
UpdateTimeCounter();
        CalculateFPS();
		process_events( );
//usleep(1000);
		draw_screen( );
	}

	return 0;
}


//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void draw_screen( void )
{
float  aspect_ratio;
 char   info_string[256];
 /////////////////////////////////
 //     RESIZE VIEWPORT         //
 /////////////////////////////////
 XGetWindowAttributes(dpy, win, &wa);
 glViewport(0, 0, wa.width, wa.height);
 aspect_ratio = (float)(wa.width) / (float)(wa.height);
//////////////////

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

/*
glMatrixMode( GL_PROJECTION );
glLoadIdentity( );
glFrustum(-1.0,1.0,-.75,.75,1.0,1000);
*/

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity( );

	xpos+=move_x;
	ypos+=move_y;
	zpos+=move_z;

	glTranslatef( xpos, ypos, zpos );
	glRotatef( angle,  1.0, 1.0, 1.0 );

	//glScalef(0.1, 0.1, 0.1);

	if( should_rotate )
	{
		if( ++angle > 360.0f )
		{
			angle = 0.0f;
		}
	}

	glPointSize(pointsize);

	glBegin(GL_LINES);
	mdel.drawcube();
	glEnd();

	glBegin(GL_POINTS);
	for(int v=0;v<mdel.nvertexes;v++)
	{
		int c;
		c=mdel.vtx[v].n;
		c=c+colorcycle;
		c%=255;
		//glColor4ub(mdel.vtx[v].r,mdel.vtx[v].g,mdel.vtx[v].b,mdel.vtx[v].a);

		if(!uselighting)
		{
			glColor4ub(cMap[c].r,cMap[c].g,cMap[c].b,mdel.vtx[v].a);
		}
		else
		{
			glColor4ub(cMap[c].r,cMap[c].g,cMap[c].b,255);
		}

		if(pointstyle==FRC_POINTS)
		{
			glVertex3f(mdel.vtx[v].x,mdel.vtx[v].y,mdel.vtx[v].z);
		}

		if(pointstyle==FRC_CUBES)
		{
			glPushMatrix();
			glTranslatef(mdel.vtx[v].x,mdel.vtx[v].y,mdel.vtx[v].z);
			auxSolidCube(1.0);
			glPopMatrix();
		}

		if(pointstyle==FRC_SPHERES)
		{
			glPushMatrix();
			glTranslatef(mdel.vtx[v].x,mdel.vtx[v].y,mdel.vtx[v].z);
			auxSolidSphere(0.5);
			glPopMatrix();
		}

	}

	if(zooming)
	{
		glBegin(GL_LINES);

		mdel.drawzoomcube();

		glEnd();
	}

	glEnd( );
	if(should_colorcycle)
	{
		colorcycle++;
		colorcycle%=255;
	}



 //     DISPLAY TIME, FPS etc.  //
 /////////////////////////////////


/*
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glOrtho(0, (float)wa.width, 0, (float)wa.height, -1., 1.);

 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();

 glColor3f(1.0, 1.0, 1.0);

 sprintf(info_string, "%4.1f seconds * %4.1f fps at %i x %i", TimeCounter, FPS, wa.width, wa.height, rot_z_vel);
 glRasterPos2i(10, 10);
 glCallLists(strlen(info_string), GL_UNSIGNED_BYTE, info_string);

 sprintf(info_string, "<up,down,left,right> rotate cube * <F1> stop rotation ");
 glRasterPos2i(10, wa.height-32);
 glCallLists(strlen(info_string), GL_UNSIGNED_BYTE, info_string);

*/

 /////////////////////////////////
 //     SWAP BUFFERS            //
 /////////////////////////////////
	//SDL_GL_SwapBuffers( );
 glXSwapBuffers(dpy, win);
}


//////////////////////////////////////////////////////////////////////

void CreateWindow(int width, int height) {

   if((dpy = XOpenDisplay(NULL)) == NULL) {
        printf("\n\tcannot connect to x server\n\n");
        exit(0);
   }

 root = DefaultRootWindow(dpy);
 
   if((vi = glXChooseVisual(dpy, 0, att)) == NULL) {
        printf("\n\tno matching visual\n\n");
        exit(0);
   }
        
   if((cmap = XCreateColormap(dpy, root, vi->visual, AllocNone)) == 0) {
        printf("\n\tcannot create colormap\n\n");
        exit(0);
   }
        
 swa.event_mask = KeyPressMask;
 swa.colormap   = cmap;
 win = XCreateWindow(dpy, root, 0, 0, width, height, 0, vi->depth, InputOutput, vi->visual, CWColormap | CWEventMask, &swa);
 XStoreName(dpy, win, "OpenGL Animation");
 XMapWindow(dpy, win);
}
////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void setup_opengl( int width, int height )
{

 char           font_string[128];
 XFontStruct    *font_struct;
 /////////////////////////////////////////////////
 //     CREATE GL CONTEXT AND MAKE IT CURRENT   //
 /////////////////////////////////////////////////
   if((glc = glXCreateContext(dpy, vi, NULL, GL_TRUE)) == NULL) {
        printf("\n\tcannot create gl context\n\n");
        exit(0);
   }

 glXMakeCurrent(dpy, win, glc);
 glEnable(GL_DEPTH_TEST);
 glClearColor(0.00, 0.00, 0.40, 1.00);
    /////////////////////////////////////////////////
    //     FIND A FONT                             //
    /////////////////////////////////////////////////
    for(int font_size = 14; font_size < 32; font_size += 2) {
//        sprintf(font_string, "-adobe-courier-*-r-normal--%i-*", font_size);
        sprintf(font_string, "*-r-normal--%i-*", font_size);
        font_struct = XLoadQueryFont(dpy, font_string);
        
        if(font_struct != NULL) {
                glXUseXFont(font_struct->fid, 32, 192, 32);          
                break;
        }
    }

	float ratio = (float) width / (float) height;

	if(uselighting)
	{
		GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
		GLfloat mat_shininess[] = { 50.0 };
		GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };

		glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
		glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
		glLightfv(GL_LIGHT0, GL_POSITION, light_position);

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_DEPTH_TEST);

		glShadeModel( GL_SMOOTH );

		glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE );
		glEnable(GL_COLOR_MATERIAL);

glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);		
		//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	//glCullFace( GL_BACK );
	//glFrontFace( GL_CCW );
	//glEnable( GL_CULL_FACE );

	glClearColor( 0, 0, 0, 0 );

	glViewport( 0, 0, width, height );

	glMatrixMode( GL_PROJECTION );

	glLoadIdentity( );

	//gluPerspective( 60.0, ratio, 1.0, 1024.0 );
	glFrustum(-1.0,1.0,-.75,.75,1.0,1000);

}


//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////////
//                              TIME COUNTER FUNCTIONS                          //
//////////////////////////////////////////////////////////////////////////////////
void InitTimeCounter() {
 gettimeofday(&tv0, NULL);
 FramesPerFPS = 30; }

void UpdateTimeCounter() {
 LastFrameTimeCounter = TimeCounter;
 gettimeofday(&tv, NULL);
 TimeCounter = (float)(tv.tv_sec-tv0.tv_sec) + 0.000001*((float)(tv.tv_usec-tv0.tv_usec));
 DT = TimeCounter - LastFrameTimeCounter;
}

void CalculateFPS() {
 Frame ++;

   if((Frame%FramesPerFPS) == 0) {
        FPS = ((float)(FramesPerFPS)) / (TimeCounter-prevTime);
        prevTime = TimeCounter;
   }
}


//                              EXIT PROGRAM                                    //
//////////////////////////////////////////////////////////////////////////////////
void ExitProgram() {
 glXMakeCurrent(dpy, None, NULL);
 glXDestroyContext(dpy, glc);
 XDestroyWindow(dpy, win);
 XCloseDisplay(dpy);
 exit(0);
}
////////////

//////////////////////////////////////////////////////////////////////
void process_events( void )
{



	while( XCheckWindowEvent(dpy, win, KeyPressMask, &xev) )
	{

		switch( xev.type )
		{
			case KeyPress:
				handle_key_down( XLookupKeysym(&xev.xkey, 0));
				break;

			/*case SDL_MOUSEMOTION:
				MouseMotionHandler(event.type, event.motion);
				break;

			case SDL_MOUSEBUTTONDOWN:
				MouseButtonHandler(event.type, event.button);
				break;

			case SDL_MOUSEBUTTONUP:
				MouseButtonHandler(event.type, event.button);
				break;

			case SDL_QUIT:
				quit_program( 0 );
				break;
*/
		}

	}

}


//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void handle_key_down( int xk )
{

	switch( xk )
	{
		case XK_Escape:
			 ExitProgram();
			break;
		case XK_space:
			should_rotate = !should_rotate;
			break;

		case XK_c:
			should_colorcycle = !should_colorcycle;
			break;

		case XK_l:
			uselighting = !uselighting;
			break;

		case XK_r:
			pointstyle++;
			pointstyle%=3;
			break;

		case XK_p:
			pointsize++;
			break;

		case XK_o:
			pointsize--;
			if(pointsize==0)pointsize=1;
			break;

		case XK_Left:
			// move_x+=0.01;
			mdel.translatezoomcube(-1.0,0.0,0.0);
			break;

		case XK_Right:
			//move_x-=0.01;
			mdel.translatezoomcube(1.0,0.0,0.0);
			break;

		case XK_Up:
			//move_y+=0.01;
			mdel.translatezoomcube(0.0,1.0,0.0);
			break;

		case XK_Down:
			//move_y-=0.01;
			mdel.translatezoomcube(0.0,-1.0,0.0);
			break;

		case XK_Home:
			//move_z+=0.01;
			mdel.translatezoomcube(0.0,0.0,1.0);
			break;

		case XK_End:
			//move_z-=0.01;
			mdel.translatezoomcube(0.0,0.0,-1.0);
			break;

		case XK_Page_Up:
			zooming = GL_TRUE;
			mdel.scalezoomcube(0.9,0.9,0.9);
			break;
		case XK_a:
			mdel.alphamode=!mdel.alphamode;
			mdel.iterate();
			break;
		case XK_j:
			mdel.julia=!mdel.julia;
			mdel.x1=-2.5;
			mdel.y1=-2.5;
			mdel.z1=-2.5;
			mdel.x2=2.5;
			mdel.y2=2.5;
			mdel.z2=2.5;

			mdel.iterate();
			break;

		case XK_Page_Down:

			mdel.scalezoomcube(1.1,1.1,1.1);

			break;

		case XK_Return:
			zooming = GL_FALSE;
			mdel.zoom();

			mdel.setcube();
			mdel.setzoomcube();
			mdel.iterate();
			break;

		case XK_F1:
			mdel.lowcolor+=1;
			if(mdel.lowcolor==256)mdel.lowcolor=255;
			mdel.iterate();
			break;

		case XK_F2:
			mdel.lowcolor-=1;
			if(mdel.lowcolor==-2)mdel.lowcolor=-1;
			mdel.iterate();
			break;

		case XK_F3:
			mdel.highcolor+=1;
			if(mdel.highcolor==256)mdel.highcolor=255;
			mdel.iterate();
			break;

		case XK_F4:
			mdel.highcolor-=1;
			if(mdel.highcolor==-2)mdel.highcolor=-1;
			mdel.iterate();
			break;

		default:
			break;
	}

}


//////////////////////////////////////////////////////////////////////
/*
void MouseButtonHandler(Uint8 type, SDL_MouseButtonEvent event)
{
	switch(event.button)
	{
		case SDL_BUTTON_LEFT:
			if (type == SDL_MOUSEBUTTONDOWN)
			{
				mouse_left = 1;
			}
			else
			{
				mouse_left = 0;
			}
			break;

		case SDL_BUTTON_RIGHT:
			if (type == SDL_MOUSEBUTTONDOWN)
			{
				mouse_right = 1;
			}
			else
			{
				mouse_right = 0;
			}
	}
}


void MouseMotionHandler(Uint8 type, SDL_MouseMotionEvent event)
{

	//xpos+=(event.xrel/50);
	//ypos+=(event.yrel/50);
}

*/

//////////////////////////////////////////////////////////////////////
void quit_program( int code )
{

//	SDL_Quit( );

	/* Exit program. */
	exit( code );
}


//////////////////////////////////////////////////////////////////////
